@@ -5,22 +5,17 @@
 # Copyright 1998-1999 - Matt Gerassimoff
 # and Ken Cox <kenstir@senteinc.com>
 #
-# $Id: gnatsweb.pl,v 2.4 1999/07/27 16:03:23 kenstir Exp $
+# $Id: gnatsweb.pl,v 1.109 1999/10/28 13:48:03 kenstir Exp $
 #
 
 #-----------------------------------------------------------------------------
 # Site-specific customization -
 #
-#     User serviceable parts - variables and subroutines.
-#
-#     We suggest you don't edit these variables here, but instead put
-#     them in a file called 'gnatsweb-site.pl' in the same directory.
+#     WE STRONGLY SUGGEST you don't edit these variables here, but instead
+#     put them in a file called 'gnatsweb-site.pl' in the same directory.
 #     That way, when a new version of gnatsweb is released, you won't
 #     need to edit them again.
 #
-#     For an example of what you can do with the site_callback
-#     subroutine, see gnatsweb-site-sente.pl.
-#
 
 # Info about your gnats host.
 $site_gnats_host = 'localhost';
@@ -29,8 +24,13 @@ $site_gnats_port = 1529;
 # Set to true if you compiled gnats with GNATS_RELEASE_BASED defined.
 $site_release_based = 0;
 
-# Name you want in the page banner.
+# Name you want in the page banner and banner color.
 $site_banner_text = 'gnatsweb';
+$site_banner_background = '#000000';
+$site_banner_foreground = '#ffffff';
+
+# Page background color -- not used unless defined.
+#$site_background = '#c0c0c0';
 
 # Program to send email notifications.
 if (-x '/usr/sbin/sendmail')
@@ -56,6 +56,8 @@ else
 #
 #     Sorry, the reasons are not documented.  Either put a call to
 #     'warn' into your gnats-site.pl file, or search this file for 'cb('.
+#     For examples of some of the things you can do with the site_callback
+#     subroutine, see gnatsweb-site-sente.pl.
 #
 # arguments:
 #     $reason - reason for the call.  Each reason is unique.
@@ -80,14 +82,18 @@ else
 # end customization
 #-----------------------------------------------------------------------------
 
-use CGI qw/:standard :netscape/;
+# Use CGI::Carp first, so that fatal errors come to the browser, including
+# those caused by old versions of CGI.pm.
 use CGI::Carp qw/fatalsToBrowser/;
+# 8/22/99 kenstir: CGI.pm-2.50's file upload is broken.
+# 9/19/99 kenstir: CGI.pm-2.55's file upload is broken.
+use CGI 2.56 qw/:standard :netscape/;
 use Socket;
 use IO::Handle;
 
-# get RCS tag as just a number
-$VERSION = '2.4';
-( $REVISION = '$Revision: 2.4 $ ' ) =~ s/.Revision: (.*) ../$1/;
+# Version number + RCS revision number
+$VERSION = '2.5';
+$REVISION = (split(/ /, '$Revision: 1.109 $ '))[1];
 
 # width of text fields
 $textwidth = 60;
@@ -345,6 +351,264 @@ sub client_cmd
   return get_reply();
 }
 
+# Return true if module MIME::Base64 is available.  If available, it's
+# loaded the first time this sub is called.
+sub can_do_mime
+{
+  return $can_do_mime if (defined($can_do_mime));
+
+  eval 'use MIME::Base64;';
+  if ($@) {
+    warn "NOTE: Can't use file upload feature without MIME::Base64 module\n";
+    $can_do_mime = 0;
+  }
+  else {
+    $can_do_mime = 1;
+  }
+  $can_do_mime;
+}
+
+# Take the file attachment's file name, and return only the tail.  Don't
+# want to store any path information, for security and clarity.  Support
+# both DOS-style and Unix-style paths here, because we have both types of
+# clients.
+sub attachment_filename_tail
+{
+  my($filename) = @_;
+  $filename =~ s,.*/,,;  # Remove leading Unix path elements.
+  $filename =~ s,.*\\,,; # Remove leading DOS path elements.
+
+  return $filename;
+}
+
+# Retrieve uploaded file attachment.  Return it as
+# ($filename, $content_type, $data).  Returns (undef,undef,undef)
+# if not present.
+#
+# See 'perldoc CGI' for details about this code.
+sub get_attachment
+{
+  my $upload_param_name = shift;
+  my $debug = 0;
+  my $filename = $q->param($upload_param_name);
+  return (undef, undef, undef) unless $filename;
+
+  # 9/6/99 kenstir: My testing reveals that if uploadInfo returns undef,
+  # then you can't read the file either.
+  warn "get_attachment: filename=$filename\n" if $debug;
+  my $hashref = $q->uploadInfo($filename);
+  if (!defined($hashref)) {
+    die "Got attachment filename ($filename) but no attachment data!  Probably this is a programming error -- the form which submitted this data must be multipart/form-data (start_multipart_form()).";
+  }
+  if ($debug) {
+    while (($k, $v) = each %$hashref) {
+      warn "get_attachment: uploadInfo($k)=$v\n";
+    }
+  }
+
+  # 9/6/99 kenstir: When testing locally on Linux, a .gz file yielded
+  # no Content-Type.  Therefore, have to assume binary.  Would like to
+  # check (-B $fh) to see if the stream is binary but that doesn't work.
+  my $ctype = $hashref->{'Content-Type'} || 'application/octet-stream';
+  warn "get_attachment: Content-Type=$ctype\n" if $debug;
+
+  my $data = '';
+  my $buf;
+  my $fh = $q->upload($upload_param_name);
+  warn "get_attachment: fh=$fh\n" if $debug;
+  while (read($fh, $buf, 1024)) {
+    $data .= $buf;
+  }
+  close $fh;
+
+  return ($filename, $ctype, $data);
+}
+
+# Retrieve uploaded file attachment, and encode it so that it's 
+# printable, for inclusion into the PR text.
+#
+# Returns the printable text representing the attachment.  Returns '' if
+# the attachment was not present.
+sub encode_attachment
+{
+  my $upload_param_name = shift;
+  my $debug = 0;
+
+  return '' unless can_do_mime();
+
+  my ($filename, $ctype, $data) = get_attachment($upload_param_name);
+  return '' unless $filename;
+
+  # Strip off path elements in $filename.
+  $filename = attachment_filename_tail($filename);
+
+  warn "encode_attachment: $filename was ", length($data), " bytes of $ctype\n"
+        if $debug;
+  my $att = '';
+
+  # Plain text is included inline; all else is encoded.
+  $att .= "Content-Type: $ctype; name=\"$filename\"\n";
+  if ($ctype eq 'text/plain') {
+    $att .= "Content-Disposition: inline; filename=\"$filename\"\n\n";
+    $att .= $data;
+  }
+  else {
+    $att .= "Content-Transfer-Encoding: base64\n";
+    $att .= "Content-Disposition: attachment; filename=\"$filename\"\n\n";
+    $att .= encode_base64($data);
+  }
+  warn "encode_attachment: done\n" if $debug;
+
+  return $att;
+}
+
+# Takes the encoded file attachment, decodes it and returns it as a hashref.
+sub decode_attachment
+{
+  my $att = shift;
+  my $debug = 0;
+  my $hash_ref = {'original_attachment' => $att};
+
+  # Split the envelope from the body.
+  my ($envelope, $body) = split(/\n\n/, $att);
+  return $hash_ref unless ($envelope && $body);
+
+  # Got the idea from this from the perldoc for split.
+  # The extra map step is the only way I could think of to strip
+  # the trailing newlines from the hash values.
+  warn "decode_attachment: envelope=>$envelope<=\n" if $debug;
+  #%$hash_ref = (USELESS_LEADING_ENTRY => split /^(\S*?):\s*/m, $envelope);
+  %$hash_ref = (map {chomp; $_;}
+               (USELESS_LEADING_ENTRY => split /^(\S*?):\s*/m, $envelope));
+  delete($$hash_ref{USELESS_LEADING_ENTRY});
+
+  # Keep the original_attachment intact.
+  $$hash_ref{'original_attachment'} = $att;
+
+  if (!$$hash_ref{'Content-Type'}
+      || !$$hash_ref{'Content-Disposition'})
+  {
+    die "Unable to parse file attachment";
+  }
+
+  # Parse filename.
+  # Note: the extra \ before the " is just so that perl-mode can parse it.
+  if ($$hash_ref{'Content-Disposition'} !~ /(\S+); filename=\"([^\"]+)\"/) {
+    die "Unable to parse file attachment Content-Disposition";
+  }
+  $$hash_ref{'filename'} = attachment_filename_tail($2);
+
+  # Decode the data if encoded.
+  if (exists($$hash_ref{'Content-Transfer-Encoding'})
+      && $$hash_ref{'Content-Transfer-Encoding'} eq 'base64')
+  {
+    $$hash_ref{'data'} = decode_base64($body);
+  }
+  else {
+    $$hash_ref{'data'} = $body;
+  }
+
+  return $hash_ref;
+}
+
+# Print file attachment browser and buttons to download the attachments.
+# Which of these appear depend on the mode.
+sub print_attachments
+{
+  my($fields_hash_ref, $mode) = @_;
+
+  return unless can_do_mime();
+
+  print "<tr><td valign=top><b>File Attachments:</b><td>";
+
+  # Add file upload button for adding new attachment.
+  if ($mode eq 'sendpr' || $mode eq 'edit') {
+    print "Add a file attachment:<br>",
+          $q->filefield(-name=>'attached_file',
+                        -size=>50);
+  }
+
+  # Print table of existing attachments.
+  # Add column with delete button in edit mode.
+  my $array_ref = $$fields_hash_ref{'attachments'};
+  my $table_rows_aref = [];
+  my $i = 0;
+  foreach $hash_ref (@$array_ref) {
+    my $size = int(length($$hash_ref{'data'}) / 1024.0);
+    $size = 1 if ($size < 1);
+    my $row_data = $q->td( [ $q->submit('cmd', "download attachment $i"),
+                             $$hash_ref{'filename'},
+                             "${size}k" ] );
+    $row_data .= $q->td($q->checkbox(-name=>'delete attachments',
+                                     -value=>$i,
+                                     -label=>"delete attachment $i"))
+          if ($mode eq 'edit');
+    push(@$table_rows_aref, $row_data);
+    $i++;
+  }
+  if (@$table_rows_aref)
+  {
+    my $header_row_data = $q->th( ['download','filename','size' ] );
+    $header_row_data .= $q->th('delete')
+          if ($mode eq 'edit');
+    print $q->table({-border=>1},
+                    $q->Tr($header_row_data),
+                    $q->Tr($table_rows_aref));
+  }
+}
+
+# The user has requested download of a particular attachment.
+# Serve it up.
+sub download_attachment
+{
+  my $attachment_number = shift;
+  my($pr) = $q->param('pr');
+  die "download_attachment called with no PR number"
+        if(!$pr);
+
+  my(%fields) = readpr($pr);
+  my $array_ref = $fields{'attachments'};
+  my $hash_ref = $$array_ref[$attachment_number];
+  print $q->header(-type => 'application/octet-stream',
+                   -content_disposition => "attachment; filename=\"$$hash_ref{'filename'}\""),
+        $$hash_ref{'data'};
+}
+
+# Add the given (gnatsweb-encoded) attachment to the %fields hash.
+sub add_encoded_attachment_to_pr
+{
+  my($fields_hash_ref, $encoded_attachment) = @_;
+  return unless $encoded_attachment;
+  my $ary_ref = $$fields_hash_ref{'attachments'} || [];
+  my $hash_ref = { 'original_attachment' => $encoded_attachment };
+  push(@$ary_ref, $hash_ref);
+  $$fields_hash_ref{'attachments'} = $ary_ref;
+}
+
+# Add the given (gnatsweb-decoded) attachment to the %fields hash.
+sub add_decoded_attachment_to_pr
+{
+  my($fields_hash_ref, $decoded_attachment_hash_ref) = @_;
+  return unless $decoded_attachment_hash_ref;
+  my $ary_ref = $$fields_hash_ref{'attachments'} || [];
+  push(@$ary_ref, $decoded_attachment_hash_ref);
+  $$fields_hash_ref{'attachments'} = $ary_ref;
+}
+
+# Remove the given attachments from the %fields hash.
+sub remove_attachments_from_pr
+{
+  my($fields_hash_ref, @attachment_numbers) = @_;
+  return unless @attachment_numbers;
+  my $ary_ref = $$fields_hash_ref{'attachments'} || [];
+  foreach my $attachment_number (@attachment_numbers)
+  {
+    # Remove the attachment be replacing it with the empty hash.
+    # The sub unparsepr skips these.
+    $$ary_ref[$attachment_number] = {};
+  }
+}
+
 sub sendpr
 {
   my $page = 'Create PR';
@@ -361,17 +625,21 @@ sub sendpr
   shift(@state);
   shift(@submitter_id);
 
-  print $q->start_form(),
+  print $q->start_multipart_form(),
+	$q->submit('cmd', 'submit'),
+	" or ",
+	$q->reset(-name=>'reset'),
+	"<hr>\n",
 	"<table>";
   print "<tr><td><b>Reporter's email:</b><td>",
         $q->textfield(-name=>'email',
                       -default=>$email,
                       -size=>$textwidth),
-	"<tr><td><b>Send a CC of this report only to:</b><td>",
-	$q->textfield(-name=>'cc',
-	              -default=>$cc,
-		      -size=>$textwidth),
-	"<tr><td><b>Others to notify<br>of updates:</b><td>",
+#	"<tr><td><b>Send a CC of this report only to:</b><td>",
+#	$q->textfield(-name=>'cc',
+#	              -default=>$cc,
+#		      -size=>$textwidth),
+	"<tr><td><b>CC these people<br>on PR status email:</b><td>",
 	$q->textfield(-name=>'X-GNATS-Notify',
 		      -size=>$textwidth),
         # a blank row, to separate header info from PR info
@@ -389,6 +657,8 @@ sub sendpr
     $default = 'medium'                   if /Priority/;
     $default = $submitter_id || 'unknown' if /Submitter-Id/;
     $default = $originator                if /Originator/;
+    $default = grep(/^unknown$/i, @category) ? "unknown" : $category[0]
+                                        if /Category/;
     $default = cb("sendpr_$lc_fieldname") || $default;
 
     if ($fieldnames{$_} & $ENUM)
@@ -408,6 +678,8 @@ sub sendpr
                          -cols=>$textwidth,
                          -rows=>$rows,
                          -default=>$default);
+      # Create file upload button after Description.
+      print_attachments(\%fields, 'sendpr') if /Description/;
     }
     else
     {
@@ -439,8 +711,10 @@ sub validate_email_field
   {
     return "$fieldname is blank";
   }
-  $email_addr = '\S+(\@\S+\.\S+)?';
-  if (!$blank && $fieldval !~ /^\s*($email_addr\s*)*$/)
+  # From rkimball@vgi.com, allows @ only if it's followed by what looks
+  # more or less like a domain name.
+  my $email_addr = '[^@\s]+(@\S+\.\S+)?';
+  if (!$blank && $fieldval !~ /^\s*($email_addr\s*)+$/)
   {
     return "'$fieldval' doesn't look like a valid email address (xxx\@xxx.xxx)";
   }
@@ -458,12 +732,14 @@ sub validate_new_pr
   # validate email fields
   $err = validate_email_field('E-mail Address', $fields{'email'}, 'required');
   push(@errors, $err) if $err;
-  $err = validate_email_field('CC', $fields{'cc'});
-  push(@errors, $err) if $err;
+#  $err = validate_email_field('CC', $fields{'cc'});
+#  push(@errors, $err) if $err;
   $err = validate_email_field('X-GNATS-Notify', $fields{'X-GNATS-Notify'});
   push(@errors, $err) if $err;
 
   # validate some other fields
+  push(@errors, "Category is blank or \"unknown\"")
+        if($fields{'Category'} =~ /^\s*$/ || $fields{'Category'} eq "unknown");
   push(@errors, "Synopsis is blank")
         if($fields{'Synopsis'} =~ /^\s*$/);
   push(@errors, "Release is blank")
@@ -507,11 +783,14 @@ sub submitnewpr
   # Supply a default value for Originator
   $fields{'Originator'} = $fields{'Originator'} || $fields{'email'};
 
+  # Handle the attached_file, if any.
+  add_encoded_attachment_to_pr(\%fields, encode_attachment('attached_file'));
+
   # Compose the message
   my $text = unparsepr('send', %fields);
   $text = <<EOT . $text;
 To: $config{'GNATS_ADDR'}
-CC: $fields{'cc'}
+CC: $fields{'X-GNATS-Notify'}
 Subject: $fields{'Synopsis'}
 From: $fields{'email'}
 Reply-To: $fields{'email'}
@@ -524,7 +803,8 @@ EOT
   if($debug)
   {
     print "<h3>debugging -- PR NOT SENT</h3>",
-          $q->pre($q->escapeHTML($text));
+          $q->pre($q->escapeHTML($text)),
+          "<hr>";
     page_end_html($page);
     return;
   }
@@ -556,6 +836,23 @@ investigate the problem.";
   page_end_html($page);
 }
 
+sub get_viewpr_url
+{
+  my $pr = shift;
+  $q->url() . "?cmd=view&database=$database&pr=$pr";
+}
+
+# Return a link which sends email regarding the current PR.
+sub get_mailto_link
+{
+  my($pr,%fields) = @_;
+  my $mailto  = $q->escape(scalar(interested_parties($pr, %fields)));
+  my $subject = $q->escape("Re: $fields{'Category'}/$pr");
+  my $body    = $q->escape(get_viewpr_url($pr));
+  return "<a href=\"mailto:$mailto?Subject=$subject&Body=$body\">"
+        . "send email to interested parties</a>\n";
+}
+
 sub view
 {
   my($viewaudit, $tmp) = @_;
@@ -579,24 +876,19 @@ sub view
   print $q->start_form(),
 	$q->hidden('pr', $pr);
 
-  # print 'edit' and 'view audit-trail' buttons as appropriate
+  # print 'edit' and 'view audit-trail' buttons as appropriate, mailto link
   print $q->submit('cmd', 'edit')             if (can_edit());
   print " or "                                if (can_edit() && !$viewaudit);
   print $q->submit('cmd', 'view audit-trail') if (!$viewaudit);
-
-  # print a convenient 'mailto' link
-  my $mailto  = $q->escape(scalar(interested_parties($pr, %fields)));
-  my $subject = $q->escape("Re: $fields{'Category'}/$pr");
-  my $body    = $q->escape(get_viewpr_url($pr));
   print " or ",
-        "<a href=\"mailto:$mailto?Subject=$subject&Body=$body\">",
-        "send email to interested parties</a>";
+        get_mailto_link($pr, %fields);
 
   print $q->hr(),
         "<table>";
   print "<tr><td><b>Reporter's email:</b><td>",
         $q->tt($fields{'Reply-To'}),
-	"<tr><td><b>Others to notify<br>of updates:</b><td>",
+#	"<tr><td><b>Others to notify<br>of updates to this PR:</b><td>",
+	"<tr><td><b>CC these people<br>on PR status email:</b><td>",
 	$q->tt($fields{'X-GNATS-Notify'}),
         # a blank row, to separate header info from PR info
         "<tr><td>&nbsp;<td>&nbsp;\n";
@@ -614,19 +906,18 @@ sub view
     }
     print "<tr><td nowrap $valign><b>$_:</b><td>",
           $q->tt($val), "\n";
+    # Print attachments after Description.
+    print_attachments(\%fields, 'view') if /Description/;
   }
   print "</table>",
         $q->hr();
 
-  # print 'edit' and 'view audit-trail' buttons as appropriate
+  # print 'edit' and 'view audit-trail' buttons as appropriate, mailto link
   print $q->submit('cmd', 'edit')             if (can_edit());
   print " or "                                if (can_edit() && !$viewaudit);
   print $q->submit('cmd', 'view audit-trail') if (!$viewaudit);
-
-  # print a convenient 'mailto' link
   print " or ",
-        "<a href=\"mailto:$mailto?Subject=$subject\">",
-        "send email to interested parties</a>";
+        get_mailto_link($pr, %fields);
 
   print $q->end_form();
 
@@ -673,10 +964,12 @@ sub edit
   shift(@state);
   shift(@submitter_id);
 
-  print $q->start_form(),
+  print $q->start_multipart_form(),
         $q->submit('cmd', 'submit edit'),
         " or ",
         $q->reset(-name=>'reset'),
+        " or ",
+        get_mailto_link($pr, %fields),
 	$q->hidden(-name=>'Editor', -value=>$user, -override=>1),
 	$q->hidden(-name=>'Last-Modified',
 		   -value=>$fields{'Last-Modified'},
@@ -689,7 +982,8 @@ sub edit
         $q->textfield(-name=>'Reply-To',
                       -default=>$fields{'Reply-To'},
                       -size=>$textwidth),
-	"<tr><td><b>Others to notify<br>of updates:</b><td>",
+#	"<tr><td><b>Others to notify<br>of updates to this PR:</b><td>",
+	"<tr><td><b>CC these people<br>on PR status email:</b><td>",
 	$q->textfield(-name=>'X-GNATS-Notify',
                       -default=>$fields{'X-GNATS-Notify'},
 		      -size=>$textwidth),
@@ -698,17 +992,17 @@ sub edit
 
   foreach (@fieldnames)
   {
-    next if ($fieldnames{$_} && $fieldnames{$_} & $EDITEXCLUDE);
+    next if ($fieldnames{$_} && ($fieldnames{$_} & $EDITEXCLUDE));
     my $lc_fieldname = field2param($_);
 
-    if ($fieldnames{$_} && $fieldnames{$_} & $ENUM)
+    if ($fieldnames{$_} && ($fieldnames{$_} & $ENUM))
     {
       print "<tr><td><b>$_:</b><td>",
             $q->popup_menu(-name=>$_,
                            -values=>\@$lc_fieldname,
                            -default=>$fields{$_});
     }
-    elsif ($fieldnames{$_} && $fieldnames{$_} & $MULTILINE)
+    elsif ($fieldnames{$_} && ($fieldnames{$_} & $MULTILINE))
     {
       my $rows = 4;
       $rows = 8 if /Description/;
@@ -718,6 +1012,8 @@ sub edit
                          -cols=>$textwidth,
                          -rows=>$rows,
                          -default=>$fields{$_});
+      # Print attachments after Description.
+      print_attachments(\%fields, 'edit') if /Description/;
     }
     else
     {
@@ -741,6 +1037,8 @@ sub edit
 	$q->submit('cmd', 'submit edit'),
 	" or ",
 	$q->reset(-name=>'reset'),
+        " or ",
+        get_mailto_link($pr, %fields),
 	$q->end_form(),
 	$q->hr();
 
@@ -752,10 +1050,25 @@ sub edit
   page_end_html($page);
 }
 
-sub get_viewpr_url
+# Print out the %fields hash for debugging.
+sub debug_print_fields
 {
-  my $pr = shift;
-  $q->url() . "?cmd=view&database=$database&pr=$pr";
+  my $fields_hash_ref = shift;
+  foreach my $f (sort keys %$fields_hash_ref)
+  {
+    print "<tr valign=top><td>$f</td><td>",
+          $q->pre($q->escapeHTML($$fields_hash_ref{$f})),
+          "</td></tr>\n";
+  }
+  my $aref = $$fields_hash_ref{'attachments'} || [];
+  my $i=0;
+  foreach my $href (@$aref) {
+    print "<tr valign=top><td>attachment $i<td>",
+          ($$href{'original_attachment'}
+           ?  $$href{'original_attachment'} : "--- empty ---");
+    $i++;
+  }
+  print "</table></font><hr>\n";
 }
 
 sub submitedit
@@ -778,6 +1091,9 @@ sub submitedit
   my $err = '';
   my $ok = 1;
 
+  # Retrieve new attachment (if any) before locking PR, in case it fails.
+  my $encoded_attachment = encode_attachment('attached_file');
+
   my(%oldfields) = lockpr($pr, $user);
   LOCKED:
   {
@@ -799,14 +1115,18 @@ sub submitedit
       $fields{$key} = $val;
     }
 
+    # Add the attached file, if any, to the new PR.
+    add_encoded_attachment_to_pr(\%fields, $encoded_attachment);
+
+    # Delete any attachments, if directed.
+    my(@deleted_attachments) = $q->param('delete attachments');
+    remove_attachments_from_pr(\%fields, @deleted_attachments);
+
     if ($debug)
     {
-      print "<h3>debugging</h3><font size=1><pre>";
-      foreach my $f (sort keys %fields)
-      {
-	printf "%-16s %s\n", $f, $q->escapeHTML($fields{$f});
-      }
-      print "</pre></font><hr>\n";
+      print "<h3>debugging -- PR edits not submitted</h3><font size=1><table>";
+      debug_print_fields(\%fields);
+      last LOCKED;
     }
 
     # TODO: fix this default; it doesn't seem right
@@ -926,12 +1246,12 @@ sub submitedit
       my($mailto);
       delete $mailto{''};
       $mailto = join(", ", sort(keys(%mailto)));
-      #print pre($q->escapeHTML("mailto->$mailto<-\n"));
+      #print $q->pre($q->escapeHTML("mailto->$mailto<-\n"));
       #last LOCKED; # debug
       if($mailto ne "" && $audittrail ne "")
       {
         # Use email address in responsible file as From, if present.
-        my $from = $responsible_address{$person} || $user;
+        my $from = $responsible_address{$user} || $user;
         if(open(MAILER, "|$site_mailer"))
 	{
 	  print MAILER "To: $mailto\n";
@@ -1214,6 +1534,22 @@ sub submitquery
   page_end_html($page);
 }
 
+# no_cookies -
+#     The client does not appear to be accepting cookies
+#     Output an appropriate error message
+#
+sub no_cookies
+{
+  my $page = 'No cookies';
+  page_start_html($page);
+
+  page_heading($page, 'Error');
+  print "<H3>Your browser does not appear to be accepting cookies.</H3>\n",
+        "gnatsweb requires cookies to keep track of your login and other\n",
+        "information.  Please turn on cookies, reload the login page,\n",
+        "and login again.\n";
+}
+
 # by_field -
 #     Sort routine called by display_query_results.
 #
@@ -1379,8 +1715,10 @@ sub display_query_results
     my($id, $cat, $syn, $conf, $sev,
        $pri, $resp, $state, $class, $sub,
        $arrival, $orig, $release, $lastmoddate, $closeddate,
-       $target, $keywords, $daterequired) = @{$_};
+       $quarter, $keywords, $daterequired) = @{$_};
     print "<td nowrap><a href=\"$sn?cmd=view&pr=$id\">$id</a>"; 
+    print " <a href=\"$sn?cmd=edit&pr=$id\"><font size=-1>edit</font></a>"
+          if can_edit();
     print "<td nowrap>$cat"                      if $fields{'category'};
     print "<td nowrap>$conf"                     if $fields{'confidential'};
     print "<td nowrap>$state[$state]"            if $fields{'state'};
@@ -1388,8 +1726,8 @@ sub display_query_results
     print "<td nowrap>$severity[$sev]"           if $fields{'severity'};
     print "<td nowrap>$priority[$pri]"           if $fields{'priority'};
     print "<td nowrap>", nonempty($release)      if $fields{'release'};
-    print "<td nowrap>", nonempty($target)       if($site_release_based
-                                                    && $fields{'target'});
+    print "<td nowrap>", nonempty($quarter)       if($site_release_based
+                                                    && $fields{'quarter'});
     print "<td>", nonempty($keywords)            if($site_release_based
                                                     && $fields{'keywords'});
     print "<td nowrap>$resp"                     if $fields{'responsible'};
@@ -1527,7 +1865,9 @@ sub submit_stored_query
   }
   else
   {
-    my $query_url = $q->script_name() . '?cmd=' . $q->escape('submit query')
+    # 9/10/99 kenstir: Must use full (not relative) URL in redirect.
+    # Patch by Elgin Lee <ehl@terisa.com>.
+    my $query_url = $q->url() . '?cmd=' . $q->escape('submit query')
           . '&' . $query_string;
     if ($debug)
     {
@@ -1605,7 +1945,7 @@ sub main_page
                 submit('cmd', 'query'));
   one_line_form('Advanced Query:',
                 submit('cmd', 'advanced query'));
-  one_line_form('Login again:',
+  one_line_form('Login Again:',
                 submit('cmd', 'login again'));
   one_line_form('Get Help:',
                 submit('cmd', 'help'));
@@ -1676,8 +2016,11 @@ sub page_start_html
     return;
   }
 
-  print $q->start_html(-title=>"$title - $site_banner_text",
-                       -bgcolor=>'#ffffff');
+  # Call start_html, with -bgcolor if we need to override that.
+  my @args = (-title=>"$title - $site_banner_text");
+  push(@args, -bgcolor=>$site_background)
+        if defined($site_background);
+  print $q->start_html(@args);
 
   # Add the page banner.  This banner is a string slammed to the right
   # of a 100% width table.  The data is a link back to the main page.
@@ -1697,7 +2040,7 @@ sub page_start_html
     # Netscape Unix
     # monospace/36pt works well.
     $style = <<END_OF_STYLE;
-      color:       white;
+      color:       $site_banner_foreground;
       font-family: monospace;
       /*font-family: lucidatypewriter, monospace;*/
       font-size:   36pt;
@@ -1708,7 +2051,7 @@ END_OF_STYLE
   {
     # monospace/28pt/bold works well in NS/Win95 (uses 'Courier New').
     $style = <<END_OF_STYLE;
-      color:       white;
+      color:       $site_banner_foreground;
       font-family: 'Courier New', monospace;
       font-size:   28pt;
       font-weight: 600;
@@ -1719,7 +2062,7 @@ END_OF_STYLE
   $row = $q->Tr($q->td({-align=>'right'},
                        $q->a({-style=>$style, -href=>$sn},
                              ' ', $site_banner_text, ' ')));
-  $banner = $q->table({-bgcolor=>'#000000', -width=>'100%',
+  $banner = $q->table({-bgcolor=>$site_banner_background, -width=>'100%',
                        -border=>0, -cellpadding=>0, -cellspacing=>0},
                       $row);
   print $banner;
@@ -1916,7 +2259,7 @@ sub initialize
   @deffields = ("category", "state", "responsible", "synopsis");
   @fields = ("category", "confidential", "state", "class",
              "severity", "priority",
-             "release", "target", "responsible", "submitter_id", "originator",
+             "release", "quarter", "responsible", "submitter_id", "originator",
              "arrival_date", "date_required",
              "last_modified", "closed_date", "synopsis");
 
@@ -1930,7 +2273,7 @@ sub initialize
     "Priority",
     "Responsible",
     "State",
-    "Target",
+    "Quarter",
     "Keywords",
     "Date-Required",
     "Class",
@@ -1961,7 +2304,7 @@ sub initialize
     "Priority"      => $ENUM,
     "Responsible"   => $ENUM | $REASONCHANGE | $SENDEXCLUDE | $AUDITINCLUDE,
     "State"         => $ENUM | $REASONCHANGE | $SENDEXCLUDE | $AUDITINCLUDE,
-    "Target"        => 0,
+    "Quarter"        => 0,
     "Keywords"      => 0,
     "Date-Required" => 0,
     "Class"         => $ENUM,
@@ -1981,6 +2324,8 @@ sub initialize
     "Unformatted"   => $MULTILINE | $SENDEXCLUDE | $EDITEXCLUDE,
   );
 
+  $attachment_delimiter = "----gnatsweb-attachment----\n";
+
   # gnatsd query commands: maps param name to gnatsd command
   %gnatsd_query = (
     "category"        => 'catg',
@@ -2002,8 +2347,8 @@ sub initialize
     "modifiedafter"   => 'maft',
     "closedbefore"    => 'cbfr',
     "closedafter"     => 'caft',
-    "target"	      => 'qrtr',
-    "keywords"	      => 'keyw',
+    "quarter"	      => 'qrtr',
+    "keywords"	      => 'kywd',
     "requiredbefore"  => 'bfor',
     "requiredafter"   => 'aftr',
   );
@@ -2011,8 +2356,8 @@ sub initialize
   # clear out some unused fields if not used
   if (!$site_release_based)
   {
-    @fields = grep(!/target|keywords|date_required/, @fields);
-    @fieldnames = grep(!/Target|Keywords|Date-Required/, @fieldnames);
+    @fields = grep(!/quarter|keywords|date_required/, @fields);
+    @fieldnames = grep(!/Quarter|Keywords|Date-Required/, @fieldnames);
   }
 
   my(@lines);
@@ -2128,6 +2473,16 @@ sub fix_email_addrs
 
 sub parsepr
 {
+  # 9/18/99 kenstir: This two-liner can almost replace the next 30 or so
+  # lines of code, but not quite.  It strips leading spaces from multiline
+  # fields.
+  #my $prtext = join("\n", @_);
+  #my(%fields) = ('envelope' => split /^>(\S*?):\s*/m, $prtext);
+  #  my $prtext = join("\n", @_);
+  #  my(%fields) = ('envelope' => split /^>(\S*?):(?: *|\n)/m, $prtext);
+
+  my $debug = 0;
+
   my($hdrmulti) = "envelope";
   my(%fields);
   foreach (@_)
@@ -2166,7 +2521,10 @@ sub parsepr
     }
 
     # Grab a few fields out of the envelope as it flies by
-    if($hdr eq "Reply-To" || $hdr eq "From" || $hdr eq "X-GNATS-Notify")
+    # 8/25/99 ehl: Grab these fields only out of the envelope, not
+    # any other multiline field.
+    if($hdrmulti eq "envelope" &&
+       ($hdr eq "Reply-To" || $hdr eq "From" || $hdr eq "X-GNATS-Notify"))
     {
       $arg = fix_email_addrs($arg);
       $fields{$hdr} = $arg;
@@ -2180,9 +2538,47 @@ sub parsepr
   $fields{'Reply-To'} = $fields{'Reply-To'} || $fields{'From'};
   delete $fields{'From'};
 
+  # Ensure that the pseudo-fields are initialized to avoid perl warnings.
+  $fields{'X-GNATS-Notify'} ||= '';
+
   # 3/30/99 kenstir: For some reason Unformatted always ends up with an
   # extra newline here.
-  $fields{'Unformatted'} =~ s/\n$//m;
+  $fields{'Unformatted'} =~ s/\n$//;
+
+  # Decode attachments stored in Unformatted field.
+  my $any_attachments = 0;
+  if (can_do_mime()) {
+    my(@attachments) = split(/$attachment_delimiter/, $fields{'Unformatted'});
+    # First element is any random text which precedes delimited attachments.
+    $fields{'Unformatted'} = shift(@attachments);
+    foreach $attachment (@attachments) {
+      warn "att=>$attachment<=\n" if $debug;
+      $any_attachments = 1;
+      add_decoded_attachment_to_pr(\%fields, decode_attachment($attachment));
+    }
+  }
+
+  if ($debug) {
+    warn "--- parsepr fields ----\n";
+    my %fields_copy = %fields;
+    foreach (@fieldnames)
+    {
+      warn "$_ =>$fields_copy{$_}<=\n";
+      delete $fields_copy{$_}
+    }
+    warn "--- parsepr pseudo-fields ----\n";
+    foreach (sort keys %fields_copy) {
+      warn "$_ =>$fields_copy{$_}<=\n";
+    }
+    warn "--- parsepr attachments ---\n";
+    my $aref = $fields{'attachments'} || [];
+    foreach $href (@$aref) {
+      warn "    ----\n";
+      while (($k,$v) = each %$href) {
+        warn "    $k =>$v<=\n";
+      }
+    }
+  }
 
   return %fields;
 }
@@ -2199,10 +2595,31 @@ sub unparsepr
 {
   my($purpose, %fields) = @_;
   my($tmp, $text);
+  my $debug = 0;
+
+  # First create or reconstruct the Unformatted field containing the
+  # attachments, if any.
+  $fields{'Unformatted'} ||= ''; # Default to empty.
+  warn "unparsepr 1 =>$fields{'Unformatted'}<=\n" if $debug;
+  my $array_ref = $fields{'attachments'};
+  foreach $hash_ref (@$array_ref) {
+    my $attachment_data = $$hash_ref{'original_attachment'};
+    # Deleted attachments leave empty hashes behind.
+    next unless defined($attachment_data);
+    $fields{'Unformatted'} .= $attachment_delimiter . $attachment_data;
+  }
+  warn "unparsepr 2 =>$fields{'Unformatted'}<=\n" if $debug;
+
+  # Reconstruct the text of the PR into $text.
   $text = $fields{'envelope'};
   foreach (@fieldnames)
   {
-    next if($purpose eq "send" && $fieldnames{$_} & $SENDEXCLUDE);
+    # Do include Unformatted field in 'send' operation, even though
+    # it's excluded.  We need it to hold the file attachment.
+    #next if($purpose eq "send" && $fieldnames{$_} & $SENDEXCLUDE);
+    next if(($purpose eq 'send')
+            && ($fieldnames{$_} & $SENDEXCLUDE)
+            && ($_ ne 'Unformatted'));
     if($fieldnames{$_} & $MULTILINE)
     {
       # Lines which begin with a '.' need to be escaped by another '.'
@@ -2261,6 +2678,7 @@ sub interested_parties
 
   my(@people);
   my $person;
+  my $list;
 
   ## Get list from MLPR command.
   #@people = client_cmd("mlpr $pr");
@@ -2269,14 +2687,18 @@ sub interested_parties
 
   # Get list of people by constructing it ourselves.
   @people = ();
-  foreach $person ($fields{'Reply-To'},
-                   $fields{'Responsible'},
-                   $fields{'X-GNATS-Notify'},
-                   $category_notify{$fields{'Category'}},
-                   $submitter_contact{$fields{'Submitter-Id'}},
-                   $submitter_notify{$fields{'Submitter-Id'}})
+  foreach $list ($fields{'Reply-To'},
+                 $fields{'Responsible'},
+                 $fields{'X-GNATS-Notify'},
+                 $category_notify{$fields{'Category'}},
+                 $submitter_contact{$fields{'Submitter-Id'}},
+                 $submitter_notify{$fields{'Submitter-Id'}})
   {
-    push(@people, $person) if $person;
+    if (defined($list)) {
+      foreach $person (split(/,/, $list)) {
+        push(@people, $person) if $person;
+      }
+    }
   }
 
   # Expand any unexpanded addresses, and build up the %addrs hash.
@@ -2312,12 +2734,12 @@ sub login
         "<table>",
         "<tr><td>User Name:<td>",
         $q->textfield(-name=>'user',
-                      -size=>15,
+                      -size=>20,
 		      -default=>$user),
         "<tr><td>Password:<td>",
         $q->password_field(-name=>'password',
                            -value=>$password,
-                           -size=>15,
+                           -size=>20,
                            -maxlength=>20),
 	"<tr><td>Database:<td>",
 	$q->popup_menu(-name=>'database',
@@ -2333,16 +2755,28 @@ sub login
 #
 # MAIN starts here:
 #
-# 3/18/99 kenstir: moved code inside gnats_main so that this code is
-# callable from gnatsweb.
 sub main
 {
   # Load gnatsweb-site.pl if present.  Die if there are errors;
   # otherwise the person who wrote gnatsweb-site.pl will never know it.
-  do 'gnatsweb-site.pl' if (-e 'gnatsweb-site.pl');
+  do './gnatsweb-site.pl' if (-e './gnatsweb-site.pl');
   die $@ if $@;
 
+  # Make sure nobody tries to swamp our server with a huge file attachment.
+  # Has to happen before 'new CGI'.
+  $CGI::POST_MAX = $site_post_max if defined($site_post_max);
+
+  # Create the query object.  Check to see if there was an error, which
+  # happens if the post exceeds POST_MAX.
   $q = new CGI;
+  if ($q->cgi_error()) {
+    print $q->header(-status=>$q->cgi_error());
+          $q->start_html('Error');
+    page_heading('Initialization failed', 'Error');
+    print $q->h3('Request not processed: ', $q->cgi_error());
+    exit();
+  }
+
   $sn = $q->script_name;
   $cmd = $q->param('cmd') || ''; # avoid perl -w warning
 
@@ -2357,13 +2791,15 @@ sub main
   }
 
   # If running a stored query, redirect the user immediately to the URL.
+  # This must happen before printing the header.
   if($cmd eq 'submit stored query')
   {
     submit_stored_query();
     exit();
   }
 
-  # Print all our cookies
+  # Debug: print all our cookies into server log.
+  #my @c;
   #$i = 0;
   #foreach my $y ($q->cookie())
   #{
@@ -2376,12 +2812,16 @@ sub main
   #warn "debug 0.5: total size of raw cookies: ", length("@c"), "\n";
 
   # Retrieve the gnatsweb cookie sent by the browser.
-  %cookie = $q->cookie('gnatsweb');
+  my %cookie = $q->cookie('gnatsweb');
+
+  # A flag to limit unnecessary cookie sending.
+  my $send_cookie = 0;
+
   $user     = $cookie{'user'};
   $password = $cookie{'password'};
   $database = $cookie{'database'};
   $email    = $cookie{'email'};
-  $cc       = $cookie{'cc'};
+#  $cc       = $cookie{'cc'};
   $submitter_id = $cookie{'Submitter-Id'};
   $originator   = $cookie{'Originator'};
   @columns      = defined($cookie{'columns'})
@@ -2391,6 +2831,20 @@ sub main
   # Upon login, store user/password/database in the gnatsweb cookie.
   if($cmd eq 'login')
   {
+    # Make sure the client is accepting and returning cookies.
+    # See also Note 9/23/99 below.
+    if(not keys %cookie)
+    {
+      print $q->header();
+      no_cookies();
+      exit();
+    }
+
+    $send_cookie = 1;
+
+    # Delete the test flag.
+    delete $cookie{'test_cookie'};
+
     $cookie{'user'}     = $user     = $q->param('user');
     $cookie{'password'} = $password = $q->param('password');
     $cookie{'database'} = $database = $q->param('database');
@@ -2401,8 +2855,10 @@ sub main
   # next time.
   if($cmd eq 'submit')
   {
+    $send_cookie = 1;
+
     $cookie{'email'}          = $email        = $q->param('email');
-    $cookie{'cc'}             = $cc           = $q->param('cc');
+#    $cookie{'cc'}             = $cc           = $q->param('cc');
     $cookie{'Submitter-Id'}   = $submitter_id = $q->param('Submitter-Id');
     $cookie{'Originator'}     = $originator   = $q->param('Originator');
     #$cookie{'Category'}    = $Category    = $q->param('Category');
@@ -2414,32 +2870,49 @@ sub main
   # Upon a 'submit query', store column display list in the cookie.
   if($cmd eq 'submit query')
   {
+    $send_cookie = 1;
     @columns = $q->param('columns');
     $cookie{'columns'} = join(' ', @columns);
   }
 
-  # Refresh the gnatsweb cookie, even if it hasn't been modified, so
-  # that it doesn't expire.
-  $new_cookie = $q->cookie(-name => 'gnatsweb', 
-                           -value => \%cookie,
-                           -expires => '+1y');
-  #warn "debug 3: new_cookie = $new_cookie\n";
+  # Construct the new cookie.  Make sure it is not empty.
+  $cookie{'test_cookie'} = 1 unless (keys %cookie);
+  my $new_cookie = $q->cookie(-name => 'gnatsweb',
+                              -value => \%cookie,
+                              -expires => '+1y');
 
-  # Serve the cookie.
-  print $q->header(-cookie=>$new_cookie);
-
-  ### Now determine what page the user wants to see.
-
-  # Return to the login page only if we haven't been there
-  # already, or if the user specifically requested to login again.
+  # Return to the login page if the user specifically requested to login
+  # again, or if we don't have the necessary login information.
+  # Note 9/23/99: Send the cookie now to test whether cookies are being
+  # accepted.  If not, it will be caught above.
   if($cmd eq 'login again' || !$user || !$password || !$database)
   {
+    print $q->header(-cookie=>$new_cookie);
     login();
     exit();
   }
 
+  # Downloading attachments must happen before printing the header,
+  # but after initialize().
+  if($cmd =~ /download attachment (\d+)/)
+  {
+    initialize();
+    download_attachment($1);
+    exit();
+  }
+
+  # Serve the header, with cookie if appropriate.
+  if ($send_cookie) {
+    print $q->header(-cookie=>$new_cookie);
+  }
+  else {
+    print $q->header();
+  }
+
+  # Login to the gnats server and set up variables.
   initialize();
 
+  # Now handle remaining commands.
   if($cmd eq 'create')
   {
     sendpr();
